from math import cos, sin, pi
from .base import Node, DisitributiveNode


def union(*args):
    return DisitributiveNode('union', args)


def hull(*args):
    return DisitributiveNode('hull', args)


def intersection(*args):
    return DisitributiveNode('intersection', args)


def difference(x, y):
    return Node('difference', [x, y])


def cube(x=0, y=0, z=0):
    result = Node('cube', None, [x, y, z])
    result.x = x
    result.y = y
    result.z = z
    return result


def cylinder(*args, **kwargs):
    return Node('cylinder', None, *args, **kwargs)


def sphere(*args, **kwargs):
    return Node('sphere', None, *args, **kwargs)


def polygon(*args, **kwargs):
    return Node('polygon', None, *args, **kwargs)


def polyhedron(points, faces=None, **kwargs):
    if faces is not None:
        kwargs['faces'] = faces
    return Node('polyhedron', None, points, **kwargs)


def sector(d=None, d1=None, d2=None, h=None, a=None, fn=None):
    if d is not None:
        d1 = d2 = d
    if fn is None:
        fn = 64
    assert(d1 is not None and d2 is not None and h is not None and a is not None)
    assert(d1 > 0 and d2 > 0 and h > 0 and a > 0 and fn > 0)
    bottom_points = [[0, 0, 0]]
    top_points = [[0, 0, h]]
    for i in range(fn + 1):
        angle = float(i) * a / fn
        angle_rad = angle * pi / 180
        x = d1 / 2 * cos(angle_rad)
        y = d1 / 2 * sin(angle_rad)
        bottom_points.append([x, y, 0])

        x = d2 / 2 * cos(angle_rad)
        y = d2 / 2 * sin(angle_rad)
        top_points.append([x, y, h])

    faces = []
    points = bottom_points + top_points
    top_start = len(bottom_points)
    for i in range(fn):
        bottom_idx = 1 + i
        top_idx = top_start + 1 + i
        faces.append([top_start, top_idx, top_idx + 1])
        faces.append([0, bottom_idx + 1, bottom_idx])

        faces.append([bottom_idx, bottom_idx + 1, top_idx + 1])
        faces.append([top_idx + 1, top_idx, bottom_idx])

    faces.append([0, 1, top_start + 1, top_start])
    faces.append([0, top_start, top_start + fn + 1, fn + 1])
    return polyhedron(points, [reversed(f) for f in faces], convexity=2)


def stl_model(filename, convexity=10):
    return Node('import', None, filename, convexity=convexity)


def text_model(txt, size=10, halign='left', valign='baseline', **kwargs):
    return Node('text', None, txt, size=size, halign=halign, valign=valign, **kwargs)



# coding: utf-8
import logging

_loggers = {}


class _LoggerProxy(object):

    def __init__(self, name):
        self._name = name
        self._logger = None

    def __getattr__(self, key):
        if self._logger is None:
            self._logger = logging.getLogger(self._name)
        return getattr(self._logger, key)


def get_logger(name):
    if name not in _loggers:
        _loggers[name] = _LoggerProxy(name)
    return _loggers[name]



import os
import sys
import time
import inspect
import argparse
import subprocess
import logging
import json
import uuid
import hashlib
from .module_watcher import ModuleWatcher
from .local_logging import get_logger

logger = get_logger(__name__)


class Project(object):
    _single_run_guard = False

    def __init__(self, name='Untitled', fa=3.0, fs=0.5):
        self._fa = fa
        self._fs = fs
        self.name = name
        self.parts = {}

    def add_part(self, name_or_method, model=None):
        method = None
        try:
            if callable(name_or_method):
                method = name_or_method
                name_or_method = name_or_method.__name__.lower().replace('_', '-')
                model = method()
            self.parts[name_or_method] = model
        except:  # noqa
            logger.exception("failed to add model")
            pass
        return method

    def get_part(self, name):
        return self.parts[name]

    def build_stl(self, args):
        self.build_scad(args)
        cache = self._read_cache(args.cache_file)
        if 'scad_cache' not in cache:
            cache['scad_cache'] = {}

        if not os.path.exists(args.stl_directory):
            os.makedirs(args.stl_directory)

        for name, _ in self.parts.items():
            logger.info('building %s.stl', name)
            scad_file_path = os.path.join(args.scad_directory, name + '.scad')
            stl_file_path = os.path.join(args.stl_directory, name + '.stl')

            if os.path.exists(stl_file_path) and not args.force:
                hc = self._get_files_hash(scad_file_path, stl_file_path)
                if cache['scad_cache'].get(scad_file_path, '') == hc:
                    continue

            command_args = [
                'openscad',
                scad_file_path,
                '-o', stl_file_path,
            ]
            subprocess.call(command_args, shell=False)
            hc = self._get_files_hash(scad_file_path, stl_file_path)
            cache['scad_cache'][scad_file_path] = hc
        self._write_cache(args.cache_file, cache)

    def build_scad(self, args):
        if not os.path.exists(args.scad_directory):
            os.makedirs(args.scad_directory)

        for name, model in self.parts.items():
            logger.info('building %s.scad', name)
            file_path = os.path.join(args.scad_directory, name + '.scad')
            with open(file_path, 'w') as fp:
                fp.write('$fa={:.4f};\n$fs={:.4f};\n'.format(self._fa, self._fs))
                fp.write(model.to_string())

    def watch(self, args):
        try:
            import __main__

            def build_scad_generator(args, script_path):
                def real_scad_generator(*args_array, **kwargs_hash):
                    command_args = [
                        __main__.__file__,
                        '--scad-directory', args.scad_directory,
                    ]
                    if args.debug:
                        command_args.append('--debug')
                    command_args.append('build-scad')
                    try:
                        subprocess.call(command_args, shell=False)
                    except OSError:
                        time.sleep(0.1)
                        subprocess.call(command_args, shell=False)

                return real_scad_generator
            callback = build_scad_generator(args, __main__.__file__)
            mw = ModuleWatcher(__main__.__file__, callback)
            try:
                callback()
                mw.start_watching()
                while True:
                    time.sleep(0.1)
            finally:
                mw.stop_watching()
        except ImportError:
            raise

    def _get_caller_module_name(self, depth=1):
        frm = inspect.stack()[depth + 1]
        mod = inspect.getmodule(frm[0])
        return mod.__name__

    def _read_cache(self, cache_file):
        result = {}
        if not os.path.exists(cache_file):
            return {}
        try:
            with open(cache_file, 'r') as fp:
                result = json.load(fp)
        except:  # noqa
            logger.error('reading cache failed', exc_info=True)
            result = {}
        return result

    def _write_cache(self, cache_file, cache):
        try:
            with open(cache_file, 'w') as fp:
                json.dump(cache, fp, ensure_ascii=False)
        except:  # noqa
            logger.error('writing cache failed', exc_info=True)
            return
        return

    def _get_files_hash(self, *filenames):
        try:
            h = hashlib.sha256()
            for filename in filenames:
                h.update(b'\0\0\0\1\0\0')
                with open(filename, "rb") as f:
                    for chunk in iter(lambda: f.read(4096), b""):  # noqa
                        h.update(chunk)
            return h.hexdigest()
        except Exception as e:  # noqa
            logger.error("hashing gone wrong %s %s", filename, e)
            return uuid.uuid4()

    def run(self):
        if Project._single_run_guard:
            return
        Project._single_run_guard = True

        parser = argparse.ArgumentParser(sys.argv[0])
        parser.add_argument('--scad-directory', type=str, help='directory to store .scad files', default='scad')
        parser.add_argument('--stl-directory', type=str, help='directory to store .stl files', default='stl')
        parser.add_argument('--cache-file', type=str, help='file to store some cahces', default='.yaost.cache')
        parser.add_argument('--force', action='store_true', help='force action', default=False)
        parser.add_argument('--debug', action='store_true', help='enable debug output', default=False)
        parser.set_defaults(func=lambda args: parser.print_help())
        subparsers = parser.add_subparsers(help='sub command help')

        watch_parser = subparsers.add_parser('watch', help='watch project and rebuild scad files')
        watch_parser.set_defaults(func=self.watch)

        build_scad_parser = subparsers.add_parser('build-scad', help='build scad files')
        build_scad_parser.set_defaults(func=self.build_scad)

        build_stl_parser = subparsers.add_parser('build-stl', help='build scad and stl files')
        build_stl_parser.set_defaults(func=self.build_stl)

        args = parser.parse_args()

        loglevel = logging.INFO
        if args.debug:
            loglevel = logging.DEBUG
        logging.basicConfig(level=loglevel, format="%(asctime)s [%(levelname)s] %(name)s: %(message)s")
        args.func(args)



# coding: utf8

from lazy import lazy
from math import acos, pi, sin

from .base import Node
from .vector import Vector


class Path(object):

    def __init__(self, points, ccw=False):
        self.points = points
        self.ccw = ccw

    @lazy
    def polygon(self):

        if len(self.points) < 2:
            return Node()

        pts = self.points
        extended_points = []
        for idx in range(len(pts)):
            pt = Vector(pts[idx])
            if idx == 0:
                v1 = Vector(pts[idx]) - Vector(pts[idx + 1])
            else:
                v1 = Vector(pts[idx - 1]) - Vector(pts[idx])

            if idx == len(pts) - 1:
                v2 = Vector(pts[idx]) - Vector(pts[idx - 1])
            else:
                v2 = Vector(pts[idx + 1]) - Vector(pts[idx])

            v1.z = 0
            v2.z = 0
            v1 = v1.normed
            v2 = v2.normed
            ptz = pt.z
            if self.ccw:
                ptz = -ptz
            sign = 1 if ptz >= 0 else -1

            dotproduct = v1.dot(v2)
            crossproduct = v1.cross(v2)
            if dotproduct < -1:
                dotproduct = -1
            elif dotproduct > 1:
                dotproduct = 1
            theta = acos(dotproduct) * 180 / pi

            if crossproduct.z * sign < 0:
                theta = 360 - theta

            while theta > 360:
                theta -= 360

            # print(v1, v2, dotproduct, theta, crossproduct)

            if theta > 180:
                if theta - 180 > 90:
                    sections = 16
                elif theta - 180 > 45:
                    sections = 8
                elif theta - 180 > 15:
                    sections = 2
                elif theta - 180 > 5:
                    sections = 1
                else:
                    sections = 0

                for sec in range(sections + 1):
                    angle = (theta - 180) * sign * sec
                    if sections > 0:
                        angle /= sections
                    vc = v1.rz(90 + angle)
                    rp = pt + vc * ptz
                    rp.z = 0
                    extended_points.append(rp.as_array)
            else:
                rp = pt + v1.rz(theta / 2 * sign) * sign * (ptz / sin(theta / 2 * pi / 180))

                rp.z = 0
                extended_points.append(rp.as_array)

        points = self.points + list(reversed(extended_points))
        points = [pt[:2] for pt in points]
        # print(points)

        return Node('polygon', None, points, faces=[idx for idx, _ in enumerate(points)], convexity=10)



import os
from setuptools import setup, find_packages

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# TODO use setuptools utils
install_requires = open(
    os.path.join(
        BASE_DIR,
        'requirements.txt'
    ),
    'r'
).read().splitlines()

with open(os.path.join(BASE_DIR, 'README.md'), 'r') as fh:
    long_description = fh.read()

test_requires = []

setup_requires = install_requires + []

setup(
    name='yaost',
    version='0.1.5',
    author='Andrey Proskurnev',
    author_email='andrey@proskurnev.ru',
    packages=find_packages(exclude=['tests']),
    install_requires=install_requires,
    tests_require=test_requires,
    setup_requires=setup_requires,
    description='Yet another python to openscad translator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    url='https://github.com/ariloulaleelay/yaost',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
)



# coding: utf-8
import copy
from lazy import lazy
from .vector import Vector


class DialectProxy(object):

    def __init__(self, node=None):
        self._node = node

    def __getattr__(self, key):
        def callback(*args, **kwargs):
            return Node(key, self._node, *args, **kwargs)
        return callback


class Node(object):

    def __init__(self, name, children, *args, **kwargs):
        if children is None:
            children = []
        elif not isinstance(children, list):
            children = [children]
        self._name = name
        self._children = children
        self.args = copy.deepcopy(args)
        self.kwargs = copy.deepcopy(kwargs)
        self.os = DialectProxy(self)

    def __argument_to_string(self, arg):
        if isinstance(arg, str):
            return '"{}"'.format(arg)

        if isinstance(arg, int):
            return '{}'.format(arg)

        if isinstance(arg, float):
            return '{:.6f}'.format(arg)

        if isinstance(arg, bool):
            return str(arg).lower()

        if isinstance(arg, list) or hasattr(arg, '__iter__'):
            return '[{}]'.format(','.join(self.__argument_to_string(v) for v in arg))

        raise Exception('Unknown argument type {} ({})'.format(type(arg), arg))

    def __magic_keys(self, k):
        if k in {'fa', 'fs', 'fn'}:
            return '$' + k
        return k

    def to_string(self):
        args = [self.__argument_to_string(v) for v in self.args]
        kwargs = [
            '{}={}'.format(self.__magic_keys(k), self.__argument_to_string(v))
            for k, v in self.kwargs.items()
        ]

        children = ''.join(c.to_string() for c in self._children)
        tail = ';'
        if children:
            tail = ''
            if len(self._children) > 1:
                children = '{' + children + '}'

        return '{}({}){}{}'.format(
            self._name,
            ','.join(args + kwargs),
            children,
            tail,
        )

    def _assert_numeric(self, *args):
        for arg in args:
            if isinstance(arg, (float, int)):
                continue
            raise Exception("%s neither float nor int" % (arg))

    @lazy
    def com(self):
        attr = 'com_for_{}'.format(self._name)
        if hasattr(Vector, attr):
            return getattr(Vector, attr)(self._children, *self.args, **self.kwargs)
        return Vector.com_for_children(self._children)

    @lazy
    def size(self):
        attr = 'size_for_{}'.format(self._name)
        if hasattr(Vector, attr):
            return getattr(Vector, attr)(self._children, *self.args, **self.kwargs)
        return Vector(0., 0., 0.)

    def t(self, x=0, y=0, z=0, **kwargs):
        if x == y == z == 0:
            return self

        if x in {'c', 'com'}:
            x = -self.com.x
        if y in {'c', 'com'}:
            y = -self.com.y
        if z in {'c', 'com'}:
            z = -self.com.z
        return TransformationNode('translate', self, [x, y, z], **kwargs)

    def r(self, x=0, y=0, z=0, xc=0, yc=0, zc=0, **kwargs):
        if x == y == z == 0:
            return self

        result = self
        if xc != 0 or yc != 0 or zc != 0:
            result = result.t(-xc, -yc, -zc)
            result = TransformationNode('rotate', result, [x, y, z], **kwargs)
            result = result.t(xc, yc, zc)
        else:
            result = TransformationNode('rotate', result, [x, y, z], **kwargs)
        return result

    def s(self, x=1.0, y=1.0, z=1.0, **kwargs):
        return TransformationNode('scale', self, [x, y, z], **kwargs)

    def m(self, x=0, y=0, z=0, xc=0, yc=0, zc=0, **kwargs):
        result = self
        if xc != 0 or yc != 0 or zc != 0:
            result = result.t(-xc, -yc, -zc)
            result = TransformationNode('mirror', result, [x, y, z], **kwargs)
            result = result.t(xc, yc, zc)
        else:
            result = TransformationNode('mirror', result, [x, y, z], **kwargs)
        return result

    def extrude(self, *args, **kwargs):
        return self.linear_extrude(*args, **kwargs)

    def linear_extrude(self, *args, **kwargs):
        return TransformationNode('linear_extrude', self, *args, **kwargs)

    def rotate_extrude(self, *args, **kwargs):
        return TransformationNode('rotate_extrude', self, *args, **kwargs)

    def tx(self, x, **kwargs):
        return self.t(x=x, **kwargs)

    def ty(self, y, **kwargs):
        return self.t(y=y, **kwargs)

    def tz(self, z, **kwargs):
        return self.t(z=z, **kwargs)

    def rx(self, x, **kwargs):
        return self.r(x=x, **kwargs)

    def ry(self, y, **kwargs):
        return self.r(y=y, **kwargs)

    def rz(self, z, **kwargs):
        return self.r(z=z, **kwargs)

    def mx(self, center=0, **kwargs):
        return self.m(x=1, xc=center, **kwargs)

    def my(self, center=0, **kwargs):
        return self.m(y=1, yc=center, **kwargs)

    def mz(self, center=0, **kwargs):
        return self.m(z=1, zc=center, **kwargs)

    def difference(self, other):
        return Node('difference', [self, other])

    def union(self, *other):
        return DisitributiveNode('union', [self] + list(other))

    def intersection(self, *other):
        return DisitributiveNode('intersection', [self] + list(other))

    def hull(self, *other):
        return DisitributiveNode('hull', [self] + list(other))

    def offset(self, *args, **kwargs):
        return Node('offset', [self], *args, **kwargs)

    def _apply_same_transformations_to(self, other_object):
        return other_object

    def same_moves(self, other_object):
        return other_object._apply_same_transformations_to(self)

    def __add__(self, other):
        return self.union(other)

    def __sub__(self, other):
        return self.difference(other)

    def __getattr__(self, key):
        if len(self._children) == 0 and key in self.kwargs:
            return self.kwargs[key]

        if len(self._children) != 1:
            raise AttributeError(key)

        return getattr(self._children[0], key)


class TransformationNode(Node):

    def _apply_same_transformations_to(self, other_object):
        return self.__class__(
            self._name,
            [c._apply_same_transformations_to(other_object) for c in self._children],
            *self.args,
            **self.kwargs,
        )

    def to_string(self):
        if not self.kwargs.get('clone', False):
            return super(TransformationNode, self).to_string()
        kwargs = copy.copy(self.kwargs)
        kwargs.pop('clone')
        clone = self.__class__(self._name, self._children, *self.args, **kwargs)
        union = DisitributiveNode('union', self._children + [clone])
        return union.to_string()


class DisitributiveNode(Node):

    def __init__(self, name, children, *args, **kwargs):
        flat_children = []
        for child in children:
            if child._name == name:
                for grandchild in child._children:
                    flat_children.append(grandchild)
            else:
                flat_children.append(child)
        super(DisitributiveNode, self).__init__(name, flat_children, *args, **kwargs)
